@@ -102,6 +102,14 @@ sub get_command ($$) {
     my ($self) = @_;
     $self->{Cmd};
 }
+sub get_fh_r ($) {
+    my ($self) = @_;
+    $self->{R};
+}
+sub get_fh_w ($) {
+    my ($self) = @_;
+    $self->{W};
+}
 
 # die due to a read error.
 #
@@ -133,6 +141,7 @@ sub expect ($&) {
     # Bind $_ for the benefit of the user's matcher.
     local $_ = readline $self->{R};
     defined && chomp or $self->readfail(__ "protocol message");
+    printdebug +($self->{Desc} // '')."<< $_\n";
     if (wantarray) {
 	my @r = &$match;
 	return @r if @r;
@@ -172,6 +181,20 @@ sub receive_data_blocks ($&) {
     }
 }
 
+# Receive exactly one data block and return its data.
+sub receive_data_block ($) {
+    my $self = shift;
+    my $ret;
+    $self->receive_data_blocks(sub {
+	if ($ret) {
+	    $self->bad("unexpected data-block");
+	} else {
+	    $ret = shift;
+	}
+    });
+    $ret
+}
+
 sub receive_file ($$) {
     my ($self, $ourfn) = @_;
     printdebug +($self->{Desc} // '')."() $ourfn\n";
@@ -187,6 +210,7 @@ sub send_raw (&$) {
     my ($self, $msg) = @_;
     my $fh = $self->{W};
     print $fh $msg or confess "$!";
+    $fh->flush or confess "$!";
 }
 
 sub send (&$) {
@@ -197,6 +221,10 @@ sub send (&$) {
 
 sub send_counted_message ($$$) {
     my ($self, $command, $message) = @_;
+    # length() returns the number of bytes only for non-UTF8.
+    # See: <https://salsa.debian.org/dgit-team/dgit/-/merge_requests/363#note_707041>
+    confess "ProtoConn cannot handle UTF8 strings"
+      if utf8::is_utf8($message);
     $self->send("$command ".length($message));
     $self->send_raw($message);
 }
